import typing

import chromadb
from chromadb import API
from chromadb.api.models.Collection import Collection
from chromadb.utils.embedding_functions import SentenceTransformerEmbeddingFunction
from loguru import logger
from pydantic import BaseModel
from pydantic_settings import BaseSettings
from tqdm import tqdm

from srctag.model import FileContext, RuntimeContext


class StorageDoc(BaseModel):
    document: str
    metadata: typing.Dict[str, str]
    id: str


class MetadataConstant(object):
    KEY_SOURCE = "source"
    KEY_COMMIT_SHA = "commit_sha"


class StorageConfig(BaseSettings):
    db_path: str = ""
    collection_name: str = "default_collection"

    # English: paraphrase-MiniLM-L6-v2
    # Multi langs: paraphrase-multilingual-MiniLM-L12-v2
    st_model_name: str = "paraphrase-MiniLM-L6-v2"


class Storage(object):
    def __init__(self, config: StorageConfig = None):
        if not config:
            config = StorageConfig()
        self.config = config

        self.chromadb: typing.Optional[API] = None
        self.chromadb_collection: typing.Optional[Collection] = None

    def init_chroma(self):
        if self.chromadb and self.chromadb_collection:
            return

        if self.config.db_path:
            self.chromadb = chromadb.PersistentClient(path=self.config.db_path)
        else:
            # by default, using in-memory db
            self.chromadb = chromadb.Client()

        self.chromadb_collection = self.chromadb.get_or_create_collection(
            self.config.collection_name,
            embedding_function=SentenceTransformerEmbeddingFunction(
                model_name=self.config.st_model_name
            ),
            # dis range: [0, 1]
            metadata={"hnsw:space": "l2"}
        )

    def process_file_ctx(self, file: FileContext, collection: Collection):
        """ can be overwritten for custom processing """

        targets = []
        for each in file.commits:
            # keep enough data in metadata for calc the final score
            item = StorageDoc(
                document=each.message,
                metadata={
                    MetadataConstant.KEY_SOURCE: file.name,
                    MetadataConstant.KEY_COMMIT_SHA: str(each.hexsha),
                },
                id=f"{file.name}|{each.hexsha}"
            )
            targets.append(item)

        for each in targets:
            collection.add(
                documents=[each.document],
                metadatas=[each.metadata],
                ids=[each.id],
            )

    def embed_file(self, file: FileContext):
        if not file.commits:
            logger.warning(f"no related commits found: {file.name}")
            return

        self.init_chroma()
        self.process_file_ctx(file, self.chromadb_collection)

    def embed_ctx(self, ctx: RuntimeContext):
        self.init_chroma()
        logger.info("start embedding source files")
        for each_file in tqdm(ctx.files.values()):
            self.embed_file(each_file)
        logger.info("embedding finished")



import os
import re
import typing
from enum import Enum

import git
from git import Repo, Commit
from loguru import logger

from pydantic_settings import BaseSettings
from tqdm import tqdm

from srctag.model import FileContext, RuntimeContext, SrcTagException


class FileLevelEnum(str, Enum):
    FILE: str = "FILE"
    DIR: str = "DIR"


class ScanRuleEnum(str, Enum):
    DFS: str = "DFS"
    BFS: str = "BFS"


class CollectorConfig(BaseSettings):
    repo_root: str = "."

    # file name include regex
    include_regex: str = ""
    include_file_list: typing.List[str] = []

    # commit msg include regex
    commit_include_regex: str = ""

    # set -1 to break the limit
    max_depth_limit: int = 16
    file_level: FileLevelEnum = FileLevelEnum.FILE

    # DFS: git log
    # BFS: walk the commits and get each diff files
    scan_rule: ScanRuleEnum = ScanRuleEnum.DFS


class Collector(object):
    def __init__(self, config: CollectorConfig = None):
        if not config:
            config = CollectorConfig()
        self.config = config

    def collect_metadata(self) -> RuntimeContext:
        exc = self._check_env()
        if exc:
            raise SrcTagException() from exc

        logger.info("git metadata collecting ...")
        ctx = RuntimeContext()
        self._collect_files(ctx)

        if self.config.scan_rule == ScanRuleEnum.DFS:
            self._collect_histories(ctx)
        else:
            self._collect_histories_globally(ctx)

        logger.info("metadata ready")
        return ctx

    def _check_env(self) -> typing.Optional[BaseException]:
        try:
            repo = git.Repo(self.config.repo_root, search_parent_directories=True)
            # if changed after search
            self.config.repo_root = repo.git_dir
        except BaseException as e:
            return e
        return None

    def _collect_files(self, ctx: RuntimeContext):
        """collect all files which tracked by git"""
        git_repo = git.Repo(self.config.repo_root)
        git_track_files = set([each[1].path for each in git_repo.index.iter_blobs()])
        if self.config.include_file_list:
            logger.info("use specific file list")
            for each in self.config.include_file_list:
                if each not in git_track_files:
                    logger.warning(f"specific file {each} not in git track, ignored")
                    continue
                ctx.files[each] = FileContext(each)
            # END file list loop
            return
        # END check file list

        include_regex = None
        if self.config.include_regex:
            include_regex = re.compile(self.config.include_regex)

        for each in git_track_files:
            if include_regex:
                if not include_regex.match(each):
                    continue

            if self.config.file_level == FileLevelEnum.FILE:
                ctx.files[each] = FileContext(each)
            elif self.config.file_level == FileLevelEnum.DIR:
                each_dir = os.path.dirname(each)
                ctx.files[each_dir] = FileContext(each_dir)
            else:
                raise SrcTagException(f"invalid file level: {self.config.file_level}")

        logger.info(f"file {len(ctx.files)} collected")

    def _collect_history(self, repo: Repo, file_path: str) -> typing.List[Commit]:
        kwargs = {
            "paths": file_path,
            "no-merges": True,
            "no-walk": True,
            "single-worktree": True,
            "max-count": self.config.max_depth_limit,
        }
        if self.config.commit_include_regex:
            kwargs["grep"] = self.config.commit_include_regex

        result = []
        for commit in repo.iter_commits(**kwargs):
            result.append(commit)
        return result

    def _collect_histories(self, ctx: RuntimeContext):
        git_repo = git.Repo(self.config.repo_root)

        for each_file, each_file_ctx in tqdm(ctx.files.items()):
            commits = self._collect_history(git_repo, each_file)
            each_file_ctx.commits = commits

    def _collect_histories_globally(self, ctx: RuntimeContext):
        git_repo = git.Repo(self.config.repo_root)

        include_regex = None
        if self.config.include_regex:
            include_regex = re.compile(self.config.include_regex)

        commit_include_regex = None
        if self.config.commit_include_regex:
            commit_include_regex = re.compile(self.config.commit_include_regex)

        kwargs = dict()
        if self.config.max_depth_limit != -1:
            kwargs["max_count"] = self.config.max_depth_limit

        for commit in tqdm(list(git_repo.iter_commits(**kwargs))):
            if commit_include_regex:
                if not commit_include_regex.match(commit.message):
                    continue

            for new_file in git_repo.git.show(commit.hexsha, name_only=True).split(
                    os.linesep
            ):
                if include_regex:
                    if not include_regex.match(new_file):
                        continue

                if self.config.file_level == FileLevelEnum.DIR:
                    new_file = os.path.dirname(new_file)

                each_file_ctx = ctx.files.get(new_file, None)
                if each_file_ctx:
                    each_file_ctx.commits.append(commit)



import typing
from collections import OrderedDict

import networkx
import networkx as nx
import numpy as np
import pandas as pd
from chromadb import QueryResult, Metadata
from loguru import logger
from pandas import Index
from pydantic_settings import BaseSettings
from tqdm import tqdm

from srctag.storage import Storage, MetadataConstant


class TagResult(object):
    def __init__(self, scores_df: pd.DataFrame):
        self.scores_df = scores_df

    def export_csv(self, path: str = "srctag-output.csv") -> None:
        logger.info(f"dump result to csv: {path}")
        self.scores_df.to_csv(path)

    def export_networkx(self) -> nx.Graph:
        df = self.scores_df.fillna(0)

        g = nx.Graph()
        for col in df.columns:
            g.add_node(col, color='lightcoral')

        for row in df.index:
            g.add_node(row, color='lightblue')

        for col in df.columns:
            for row in df.index:
                g.add_edge(col, row, weight=df.loc[row, col])

        return g

    def export_dot(self, path: str):
        graph = self.export_networkx()
        logger.info(f"dump result to dot: {path}")
        networkx.drawing.nx_pydot.write_dot(graph, path)

    @classmethod
    def import_csv(cls, path: str) -> "TagResult":
        scores_df = pd.read_csv(path, index_col=0)
        return TagResult(scores_df=scores_df)

    def tags(self) -> Index:
        return self.scores_df.columns

    def files(self) -> Index:
        return self.scores_df.index

    def tags_by_file(self, file_name: str) -> typing.Optional[pd.Series]:
        if file_name not in self.scores_df.index:
            return None
        return self.scores_df.loc[file_name].sort_values(ascending=False)

    def files_by_tag(self, tag_name: str) -> typing.Optional[pd.Series]:
        if tag_name not in self.scores_df.columns:
            return None
        return self.scores_df.loc[:, tag_name].sort_values(ascending=False)

    def top_n_tags(self, file_name: str, n: int) -> typing.List[str]:
        return self.tags_by_file(file_name).nlargest(n).index.tolist()

    def top_n_files(self, tag_name: str, n: int) -> typing.List[str]:
        return self.files_by_tag(tag_name).nlargest(n).index.tolist()

    def normalize_score(self, origin: float) -> float:
        return origin / len(self.files())


class TaggerConfig(BaseSettings):
    tags: typing.Set[str] = set()

    # eg: 0.3 == choosing the closest 30% results
    n_percent: float = 0.3

    # use self.optimize or not
    optimize: bool = False

    # normalization or rank
    normalize: bool = True


class Tagger(object):
    """load feature map, search related files, and tag them"""

    def __init__(self, config: TaggerConfig = None):
        if not config:
            config = TaggerConfig()
        self.config = config

    def tag(self, storage: Storage) -> TagResult:
        storage.init_chroma()
        doc_count = storage.chromadb_collection.count()
        n_results = int(doc_count * self.config.n_percent)

        logger.info(f"start tagging source files ...")

        tag_results = []
        for each_tag in tqdm(self.config.tags):
            query_result: QueryResult = storage.chromadb_collection.query(
                query_texts=each_tag,
                n_results=n_results,
                include=["metadatas", "distances"],
            )

            metadatas: typing.List[Metadata] = query_result["metadatas"][0]
            # https://github.com/langchain-ai/langchain/blob/master/libs/langchain/langchain/vectorstores/chroma.py
            # https://stats.stackexchange.com/questions/158279/how-i-can-convert-distance-euclidean-to-similarity-score
            distances: typing.List[float] = query_result["distances"][0]
            normalized_scores = [
                1.0 / (1.0 + x) for x in distances
            ]

            for each_metadata, each_score in zip(metadatas, normalized_scores):
                each_file_name = each_metadata[MetadataConstant.KEY_SOURCE]
                tag_results.append((each_tag, each_file_name, each_score))
            # END file loop
        # END tag loop

        ret = dict()
        for each_tag, each_file_name, each_score in tag_results:
            if each_file_name not in ret:
                # has not been touched by other tags
                # the score order is decreasing
                ret[each_file_name] = OrderedDict()
            each_file_tag_result = ret[each_file_name]

            if each_tag not in each_file_tag_result:
                each_file_tag_result[each_tag] = each_score
            else:
                # has been touched by other commits, merge
                each_file_tag_result[each_tag] += each_score
        # END tag_results

        scores_df = pd.DataFrame.from_dict(ret, orient="index")
        if self.config.optimize:
            scores_df = self.optimize(scores_df)

        # convert score matrix into rank (use reversed rank as score). because:
        # 1. score/distance is meaningless to users
        # 2. can not be evaluated both rows and cols
        scores_df = scores_df.rank(axis=0, method='min')

        if self.config.normalize:
            scores_df = (scores_df - scores_df.min()) / (scores_df.max() - scores_df.min())

        logger.info(f"tag finished")
        return TagResult(scores_df=scores_df)

    def optimize(self, df: pd.DataFrame) -> pd.DataFrame:
        scale_factor = 2.0
        df = np.exp(df * scale_factor)

        # reduce the impacts of common files
        row_variances = df.var(axis=1)
        max_variance = row_variances.max()
        weights = 1.0 - row_variances / max_variance
        df = df.multiply(weights, axis=0)
        return df



import chromadb
import click

from chromadb.utils.embedding_functions import SentenceTransformerEmbeddingFunction
from loguru import logger

from srctag.collector import Collector, FileLevelEnum
from srctag.storage import Storage
from srctag.tagger import Tagger


@click.group()
def cli():
    pass


@cli.command()
def prepare():
    """ usually used for pre-downloading sentence-transformer models """

    logger.info("Start checking env. It may takes a few minutes for downloading models ...")
    # try to embed
    chromadb_cli = chromadb.Client()
    collection = chromadb_cli.get_or_create_collection(
        "testonly",
        embedding_function=SentenceTransformerEmbeddingFunction(),
    )
    collection.add(
        documents=["doc"],
        ids=["id"],
    )
    assert collection.count() == 1
    click.echo("ok.")


@cli.command()
@click.option("--repo-root", default=".", help="Repository root directory")
@click.option("--max-depth-limit", default=-1, help="Maximum depth limit")
@click.option("--include-regex", default="", help="File include regex pattern")
@click.option("--tags-file", type=click.File("r"), default="./srctag.txt", help="Path to a text file containing tags")
@click.option("--output-path", default="", help="Output file path for CSV")
@click.option("--file-level", default=FileLevelEnum.FILE.value, help="Scan file level, FILE or DIR, default to FILE")
@click.option("--st-model", default="", help="Sentence Transformer Model")
@click.option("--commit-include-regex", default="", help="Commit message include regex pattern")
def tag(repo_root, max_depth_limit, include_regex, tags_file, output_path, file_level, st_model, commit_include_regex):
    """ tag your repo """
    collector = Collector()
    collector.config.repo_root = repo_root
    collector.config.max_depth_limit = max_depth_limit
    collector.config.include_regex = include_regex
    collector.config.file_level = file_level
    collector.config.commit_include_regex = commit_include_regex

    ctx = collector.collect_metadata()
    storage = Storage()
    if st_model:
        storage.config.st_model_name = st_model
    storage.embed_ctx(ctx)
    tagger = Tagger()

    assert tags_file, "no tag file provided"
    tags = [each.strip() for each in tags_file.read().splitlines()]

    tagger.config.tags = tags
    tag_dict = tagger.tag(storage)

    if output_path:
        tag_dict.export_csv(path=output_path)
    else:
        tag_dict.export_csv()


if __name__ == '__main__':
    cli()



import typing

from git import Commit


class FileContext(object):
    def __init__(self, name: str):
        self.name: str = name
        self.commits: typing.List[Commit] = []


class RuntimeContext(object):
    def __init__(self):
        self.files: typing.Dict[str, FileContext] = dict()


class SrcTagException(BaseException):
    pass
